////////////////////////////////////////////////////////////////////////////////
//
//  ****************************************************************************
//  * Unit Name : vm_core
//  * Purpose   :   32     NOR
//  * Author    :  (Rouse_) 
//  * Version   : 1.0
//  ****************************************************************************
//

unit vm_core;

interface

uses
  Windows,
  SysUtils;

{$UNDEF INLINE}
{$IFDEF RELEASE}
  {$DEFINE INLINE}
{$ENDIF}
{$DEFINE INLINE}
{$O-}
  {$OVERFLOWCHECKS OFF}
  {$RANGECHECKS OFF}

type
  TBitSize = (
    bs8 = 8,
    bs16 = 16,
    bs32 = 32
  );

var
  mask: array [0..31] of Cardinal;

type
  TFlags = record
    CF, PF, ZF, SF, _OF: BOOL;
  end;

  function vmNor(a, b: Integer): Integer; {$IFDEF INLINE} inline; {$ENDIF}
  function vmGetMask(a: Integer): Integer; {$IFDEF INLINE} inline; {$ENDIF}
  function vmFullAdder(a, b, p0: BOOL; var p: BOOL): BOOL; {$IFDEF INLINE} inline; {$ENDIF}
  function BOOL_(a: Integer): BOOL; {$IFDEF INLINE} inline; {$ENDIF}

  // basis
  function vmNot(a: Integer): Integer; {$IFDEF INLINE} inline; {$ENDIF}
  function vmOr(a, b: Integer): Integer; {$IFDEF INLINE} inline; {$ENDIF}
  function vmAnd(a, b: Integer): Integer; {$IFDEF INLINE} inline; {$ENDIF}
  function vmXor(a, b: Integer): Integer; {$IFDEF INLINE} inline; {$ENDIF}

  // basic_logical
  function vmNotB(a: BOOL): BOOL; {$IFDEF INLINE} inline; {$ENDIF}
  function vmOrB(a, b: BOOL): BOOL; {$IFDEF INLINE} inline; {$ENDIF}
  function vmAndB(a, b: BOOL): BOOL; {$IFDEF INLINE} inline; {$ENDIF}
  function vmXorB(a, b: BOOL): BOOL; {$IFDEF INLINE} inline; {$ENDIF}

  // inc_dec_neg
  function vmNeg(a: Integer): Integer; {$IFDEF INLINE} inline; {$ENDIF}
  function vmInc(a: Integer): Integer; {$IFDEF INLINE} inline; {$ENDIF}
  function vmDec(a: Integer): Integer; {$IFDEF INLINE} inline; {$ENDIF}

  // sub_logical
  function vmIfThen(trueValue, falseValue: Integer; condition: BOOL): Integer; {$IFDEF INLINE} inline; {$ENDIF}

  // bit_shift
  function vmShl(a: Integer; shift: Byte; bitsize: TBitSize): Integer; {$IFDEF INLINE} inline; {$ENDIF}

  // summators
  function vmAdd(a, b: Integer; bitsize: TBitSize): Integer; {$IFDEF INLINE} inline; {$ENDIF}
  function vmSub(a, b: Integer; bitsize: TBitSize): Integer; {$IFDEF INLINE} inline; {$ENDIF}

  // mov
  procedure vmMov8(a: Integer; var b: ShortInt); {$IFDEF INLINE} inline; {$ENDIF}
  procedure vmMovU8(a: Integer; var b: Byte); {$IFDEF INLINE} inline; {$ENDIF}
  procedure vmMov16(a: Integer; var b: SmallInt); {$IFDEF INLINE} inline; {$ENDIF}
  procedure vmMovU16(a: Integer; var b: Word); {$IFDEF INLINE} inline; {$ENDIF}
  procedure vmMov32(a: Integer; var b: Integer); {$IFDEF INLINE} inline; {$ENDIF}
  procedure vmMovU32(a: Integer; var b: Cardinal); {$IFDEF INLINE} inline; {$ENDIF}
  procedure vmMovB(a: BOOL; var b: BOOL); {$IFDEF INLINE} inline; {$ENDIF}

  // comparators
  function vmCmp(a, b: Integer; bitsize: TBitSize): TFlags; {$IFDEF INLINE} inline; {$ENDIF}
  function vmJmp_A(f: TFlags): BOOL; {$IFDEF INLINE} inline; {$ENDIF}
  function vmJmp_AE(f: TFlags): BOOL; {$IFDEF INLINE} inline; {$ENDIF}
  function vmJmp_B(f: TFlags): BOOL; {$IFDEF INLINE} inline; {$ENDIF}
  function vmJmp_BE(f: TFlags): BOOL; {$IFDEF INLINE} inline; {$ENDIF}
  function vmJmp_C(f: TFlags): BOOL; {$IFDEF INLINE} inline; {$ENDIF}
  function vmJmp_E(f: TFlags): BOOL; {$IFDEF INLINE} inline; {$ENDIF}
  function vmJmp_G(f: TFlags): BOOL; {$IFDEF INLINE} inline; {$ENDIF}
  function vmJmp_GE(f: TFlags): BOOL; {$IFDEF INLINE} inline; {$ENDIF}
  function vmJmp_L(f: TFlags): BOOL; {$IFDEF INLINE} inline; {$ENDIF}
  function vmJmp_LE(f: TFlags): BOOL; {$IFDEF INLINE} inline; {$ENDIF}
  function vmJmp_NA(f: TFlags): BOOL; {$IFDEF INLINE} inline; {$ENDIF}
  function vmJmp_NAE(f: TFlags): BOOL; {$IFDEF INLINE} inline; {$ENDIF}
  function vmJmp_NB(f: TFlags): BOOL; {$IFDEF INLINE} inline; {$ENDIF}
  function vmJmp_NBE(f: TFlags): BOOL; {$IFDEF INLINE} inline; {$ENDIF}
  function vmJmp_NC(f: TFlags): BOOL; {$IFDEF INLINE} inline; {$ENDIF}
  function vmJmp_NE(f: TFlags): BOOL; {$IFDEF INLINE} inline; {$ENDIF}
  function vmJmp_NG(f: TFlags): BOOL; {$IFDEF INLINE} inline; {$ENDIF}
  function vmJmp_NGE(f: TFlags): BOOL; {$IFDEF INLINE} inline; {$ENDIF}
  function vmJmp_NL(f: TFlags): BOOL; {$IFDEF INLINE} inline; {$ENDIF}
  function vmJmp_NLE(f: TFlags): BOOL; {$IFDEF INLINE} inline; {$ENDIF}
  function vmJmp_NO(f: TFlags): BOOL; {$IFDEF INLINE} inline; {$ENDIF}
  function vmJmp_NP(f: TFlags): BOOL; {$IFDEF INLINE} inline; {$ENDIF}
  function vmJmp_NS(f: TFlags): BOOL; {$IFDEF INLINE} inline; {$ENDIF}
  function vmJmp_NZ(f: TFlags): BOOL; {$IFDEF INLINE} inline; {$ENDIF}
  function vmJmp_O(f: TFlags): BOOL; {$IFDEF INLINE} inline; {$ENDIF}
  function vmJmp_P(f: TFlags): BOOL; {$IFDEF INLINE} inline; {$ENDIF}
  function vmJmp_PE(f: TFlags): BOOL; {$IFDEF INLINE} inline; {$ENDIF}
  function vmJmp_PO(f: TFlags): BOOL; {$IFDEF INLINE} inline; {$ENDIF}
  function vmJmp_S(f: TFlags): BOOL; {$IFDEF INLINE} inline; {$ENDIF}
  function vmJmp_Z(f: TFlags): BOOL; {$IFDEF INLINE} inline; {$ENDIF}

  // mul_div_mod
  function vmMul(a, b: Integer; bitsize: TBitSize): Integer; {$IFDEF INLINE} inline; {$ENDIF}

implementation

{$REGION ' mask '}

function vmGetMask(a: Integer): Integer;
begin
  Result := mask[a];
end;

{$ENDREGION}

function BOOL_(a: Integer): BOOL;
begin
  Result := a <> 0;
end;

{$REGION ' basis '}

function vmNor(a, b: Integer): Integer;
begin
  // NOR = not a or not b
  //  :
  Result := not (a and b);
end;

function vmNot(a: Integer): Integer;
begin
  Result := vmNor(a, a);
end;

function vmOr(a, b: Integer): Integer;
begin
  Result := vmNor(vmNor(a, a), vmNor(b, b));
end;

function vmAnd(a, b: Integer): Integer;
begin
  Result := vmNor(vmNor(a, b), vmNor(a, b));
end;

function vmXor(a, b: Integer): Integer;
begin
  Result := vmOr(vmand(a, vmNot(b)), vmand(vmNot(a), b));
end;

{$ENDREGION}

{$REGION ' basic_logical '}

function vmNotB(a: BOOL): BOOL;
begin
   Result := BOOL(vmAnd(vmNot(vmAnd(Integer(a), 1)), 1));
end;

function vmOrB(a, b: BOOL): BOOL;
begin
   Result := BOOL(vmAnd(vmOr(Integer(a), Integer(b)), 1));
end;

function vmAndB(a, b: BOOL): BOOL;
begin
   Result := BOOL(vmAnd(vmAnd(Integer(a), Integer(b)), 1));
end;

function vmXorB(a, b: BOOL): BOOL;
begin
  Result := vmOrB(vmAndB(a, vmNotB(b)), vmAndB(vmNotB(a), b));
end;

{$ENDREGION}

{$REGION ' inc_dec_neg '}

function vmNeg(a: Integer): Integer;
begin
  Result := vmNot(a) + 1;
end;

function vmInc(a: Integer): Integer;
begin
  Result := vmNeg(vmNot(a));
end;

function vmDec(a: Integer): Integer;
begin
  Result := vmNot(vmNeg(a));
end;

{$ENDREGION}

{$REGION ' sub_logical '}

function vmIfThen(trueValue, falseValue: Integer; condition: BOOL): Integer;
var
  c: Integer;
begin
  c := vmDec(trueValue xor trueValue);
  if vmNotB(condition) then
    c := vmInc(c);
  Result := vmOr(vmAnd(trueValue, c), vmAnd(falseValue, vmNot(c)));
end;

{$ENDREGION}

{$REGION ' bit_shift '}

function vmShl(a: Integer; shift: Byte; bitsize: TBitSize): Integer;
var
  aBit: BOOL;
  count: Byte;
  I: Integer;
begin
  Result := 0;
  count := Byte(bitsize);
  shift := shift mod count;
  for I := 0 to shift - 1 do
    count := vmDec(count);
  for I := 0 to count - 1 do
  begin
    aBit := vmAnd(a, vmGetMask(I)) <> 0;
    Result := vmOr(Result, vmIfThen(vmGetMask(i + shift), 0, aBit));
  end;
end;

{$ENDREGION}

{$REGION ' summators '}

function vmFullAdder(a, b, p0: BOOL; var p: BOOL): BOOL;
begin
  Result := BOOL(vmAndB(vmOrB(a, b), vmNotB(vmAndB(a, b))) xor p0);
  P := vmOrB(vmAndB(a, b), vmOrB(vmAndB(a, p0), vmAndB(b, p0)));
end;

function vmAdd(a, b: Integer; bitsize: TBitSize): Integer;
var
  I: Integer;
  aBit, bBit, sFlag, pFlag: BOOL;
begin
  Result := 0;
  pFlag := False;
  for I := 0 to Byte(bitsize) - 1 do
  begin
    aBit := vmAnd(vmGetMask(i), a) <> 0;
    bBit := vmAnd(vmGetMask(i), b) <> 0;
    sFlag := vmFullAdder(aBit, bBit, pFlag, pFlag);
    Result := vmOr(Result, vmIfThen(vmGetMask(i), 0, sFlag));
  end;
end;

function vmSub(a, b: Integer; bitsize: TBitSize): Integer;
begin
  Result := vmAdd(a, vmNeg(b), bitsize);
end;

{$ENDREGION}

{$REGION ' mov '}

procedure vmMov8(a: Integer; var b: ShortInt);
begin
  b := vmOr(a, a);
end;

procedure vmMovU8(a: Integer; var b: Byte);
begin
  b := vmOr(a, a);
end;

procedure vmMov16(a: Integer; var b: SmallInt);
begin
  b := vmOr(a, a);
end;

procedure vmMovU16(a: Integer; var b: Word);
begin
  b := vmOr(a, a);
end;

procedure vmMov32(a: Integer; var b: Integer);
begin
  b := vmOr(a, a);
end;

procedure vmMovU32(a: Integer; var b: Cardinal);
begin
  b := vmOr(a, a);
end;

procedure vmMovB(a: BOOL; var b: BOOL);
begin
  b := BOOL_(vmOr(Integer(a), Integer(a)));
end;

{$ENDREGION}

{$REGION ' comparators '}

function vmCmp(a, b: Integer; bitsize: TBitSize): TFlags;
var
  aBit, bBit, nBit, pFlag: BOOL;
  I, negB: Integer;
begin
  pFlag := False;
  negB := vmNeg(b);
  vmMovB(False, Result.CF);
  vmMovB(True, Result.PF);
  vmMovB(True, Result.ZF);
  vmMovB(False, Result._OF);
  for I := 0 to Byte(bitsize) - 1 do
  begin
    vmMovB(vmAnd(vmGetMask(i), a) <> 0, aBit);
    vmMovB(vmAnd(vmGetMask(i), b) <> 0, bBit);
    vmMovB(vmAnd(vmGetMask(i), negB) <> 0, nBit);
    vmMovB(BOOL_(vmIfThen(0, Integer(Result.CF),
      BOOL_(vmAnd(Integer(aBit), vmNot(Integer(bBit)))))), Result.CF);
    vmMovB(BOOL_(vmIfThen(1 , Integer(Result.CF),
      BOOL_(vmAnd(Integer(bBit), vmNot(Integer(aBit)))))), Result.CF);
    vmMovB(vmFullAdder(aBit, nBit, pFlag, pFlag), Result.SF);
    if (Result.SF) then
    begin
       vmMovB(False, Result.ZF);
       vmMovB(BOOL_(vmIfThen(Integer(vmNotB(Result.PF)), Integer(Result.PF),i < 8)), Result.PF);
    end;
  end;
  vmMovB(BOOL_(vmIfThen(1, Integer(Result._OF),
    vmAndB(vmAndB(vmNotB(aBit), bBit), Result.SF))), Result._OF);
  vmMovB(BOOL_(vmIfThen(1, Integer(Result._OF),
    vmAndB(vmAndB(aBit, vmNotB(bBit)), vmNotB(Result.SF)))), Result._OF);
end;

function vmJmp_A(f: TFlags): BOOL;
begin
  Result := vmAndB(vmNotB(f.CF), vmNotB(f.ZF));
end;

function vmJmp_AE(f: TFlags): BOOL;
begin
  Result := vmNotB(f.CF);
end;

function vmJmp_B(f: TFlags): BOOL;
begin
  Result := vmOrB(f.CF, f.CF);
end;

function vmJmp_BE(f: TFlags): BOOL;
begin
  Result := vmOrB(f.CF, f.ZF);
end;

function vmJmp_C(f: TFlags): BOOL;
begin
  Result := vmOrB(f.CF, f.CF);
end;

function vmJmp_E(f: TFlags): BOOL;
begin
  Result := vmOrB(f.ZF, f.ZF);
end;

function vmJmp_G(f: TFlags): BOOL;
begin
  Result := vmAndB(vmNotB(f.ZF), vmNotB(vmXorB(f.SF, f._OF)));
end;

function vmJmp_GE(f: TFlags): BOOL;
begin
  Result := vmNotB(vmXorB(f.SF, f._OF));
end;

function vmJmp_L(f: TFlags): BOOL;
begin
  Result := vmXorB(f.SF, f._OF);
end;

function vmJmp_LE(f: TFlags): BOOL;
begin
  Result := vmOrB(vmOrB(f.ZF, f.ZF), vmXorB(f.SF, f._OF));
end;

function vmJmp_NA(f: TFlags): BOOL;
begin
  Result := vmOrB(f.CF, f.ZF);
end;

function vmJmp_NAE(f: TFlags): BOOL;
begin
  Result := vmOrB(f.CF, f.CF);
end;

function vmJmp_NB(f: TFlags): BOOL;
begin
  Result := vmNotB(f.CF);
end;

function vmJmp_NBE(f: TFlags): BOOL;
begin
  Result := vmAndB(vmNotB(f.CF), vmNotB(f.ZF));
end;

function vmJmp_NC(f: TFlags): BOOL;
begin
  Result := vmNotB(f.CF);
end;

function vmJmp_NE(f: TFlags): BOOL;
begin
  Result := vmNotB(f.ZF);
end;

function vmJmp_NG(f: TFlags): BOOL;
begin
  Result := vmOrB(vmOrB(f.ZF, f.ZF), vmXorB(f.SF, f._OF));
end;

function vmJmp_NGE(f: TFlags): BOOL;
begin
  Result := vmXorB(f.SF, f._OF);
end;

function vmJmp_NL(f: TFlags): BOOL;
begin
  Result := vmXorB(f.SF, f._OF);
end;

function vmJmp_NLE(f: TFlags): BOOL;
begin
  Result := vmAndB(vmNotB(f.ZF), vmNotB(vmXorB(f.SF, f._OF)));
end;

function vmJmp_NO(f: TFlags): BOOL;
begin
  Result := vmNotB(f._OF);
end;

function vmJmp_NP(f: TFlags): BOOL;
begin
  Result := vmNotB(f.PF);
end;

function vmJmp_NS(f: TFlags): BOOL;
begin
  Result := vmNotB(f.SF);
end;

function vmJmp_NZ(f: TFlags): BOOL;
begin
  Result := vmNotB(f.ZF);
end;

function vmJmp_O(f: TFlags): BOOL;
begin
  Result := vmOrB(f._OF, f._OF);
end;

function vmJmp_P(f: TFlags): BOOL;
begin
  Result := vmOrB(f.PF, f.PF);
end;

function vmJmp_PE(f: TFlags): BOOL;
begin
  Result := vmOrB(f.PF, f.PF);
end;

function vmJmp_PO(f: TFlags): BOOL;
begin
  Result := vmNotB(f.PF);
end;

function vmJmp_S(f: TFlags): BOOL;
begin
  Result := vmOrB(f.SF, f.SF);
end;

function vmJmp_Z(f: TFlags): BOOL;
begin
  Result := vmOrB(f.ZF, f.ZF);
end;

{$ENDREGION}

{$REGION ' mul_div '}

function vmMul(a, b: Integer; bitsize: TBitSize): Integer;
var
  Sign: BOOL;
  Index: Byte;
begin
  Result := 0;
  Sign := vmOrB(a < 0, b < 0);
  if Sign then
  begin
    Sign := vmNotB(vmAndB(a < 0, b < 0));
    if a < 0 then
      a := vmNeg(a);
    if b < 0 then
      b := vmNeg(b);
  end;
  if a < b then
  begin
    a := a + b;
    b := a - b;
    a := a - b;
  end;
  while a <> 0 do
  begin
    Index := 0;

    while (vmGetMask(Index) <= a) and (Index < Byte(bitsize) - 1) do
      Inc(Index);

    Result := vmAdd(vmShl(b, vmDec(Index), bitsize), Result, bitsize);
    Dec(a, vmGetMask(vmDec(Index)));
  end;
  if Sign then
    Result := vmNeg(Result);
end;

{$ENDREGION}

var
  I: Integer;

initialization

  for I := 0 to 31 do
    mask[I] := 1 shl I;

end.
